HEADER = {
    "authority": "www.landwatch.com",
    "accept": "*/*",
    "accept-language": "en-US,en;q=0.7",
    "referer": "https://www.landwatch.com/tennessee-land-for-sale/chattanooga",
    "sec-fetch-dest": "empty",
    "sec-fetch-mode": "cors",
    "sec-fetch-site": "same-origin",
    "sec-gpc": "1",
    "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 Safari/537.36"
}

ROOT_URL = "https://www.landwatch.com/api/property/search/1113"
